# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2022-2025, by Samuel Williams.

require_relative "rack/version"
require_relative "rack/adapter"
require_relative "rack/request"
require_relative "rack/response"

# @namespace
module Protocol
	# @namespace
	module Rack
	end
end
